

import numpy as np
import scipy.ndimage as ndi
import scipy.stats as stats


def nanmean(x):
    n = np.sum(np.isfinite(x))
    return np.nanmean(x) if n>0 else np.nan


def gaussian_kernel(sigma, size=11):
    """
    Create a gaussian kernel.
    """
    if size%2 == 0:
        raise ValueError('kernel size needs to be an odd integer')
        
    f = np.zeros([size,size])
    f[size//2, size//2] = 1.0
    g = ndi.gaussian_filter(f, sigma=sigma)
    return g


def weighted_mean(x, kernel):
    """
    Computed weighted mean.
    """
    valids = np.isfinite(x)
    
    if np.any(valids):
        kernel = kernel[valids]
        kernel = kernel / kernel.sum()

        return np.sum(x[valids] * kernel)
    else:
        return np.nan


def weighted_mean_var(x, kernel, variance):
    """
    Compute variance reduction of weighted mean.
    """
    valids = np.isfinite(x)
    
    if np.any(valids):
        kernel = kernel[valids]
        kernel = kernel / kernel.sum()

        return variance * np.sum(kernel**2)
    else:
        return np.nan

   
    
    
    
    
    
    

def do_test(mean_s, mean_b, var_rand, var_sys, size, q, variance=None, dtype='f4'):
    """
    mean_s:    mean of sample
    mean_b:    mean of background
    variance:  estimated local variance
    
    var_rand:  random variance of sample
    var_sys:   systematic variance of sample
    size:      size of sample
    q:         threshold
    """
    mean_s = np.array(mean_s)
    size = np.array(size)

    if np.ndim(mean_b) == 0:
        mean_b = np.full(mean_s.shape, mean_b)

    if variance is None:
        z_values = np.full(mean_s.shape, np.nan)
        m = size > 0

        z_values[m] = mean_s[m] - mean_b[m]
        z_values[m] /= np.sqrt(var_rand / size[m] + var_sys)
    else:
        z_values = (mean_s - mean_b) / np.sqrt(variance)

    return z_values.astype(dtype), z_values > stats.norm.ppf(q)


def welch_satterthwaite_equation(s1, s2, n1, n2):
    """ Estimate degree of freedom. """
    a = (s1**2 / n1 + s2**2 / n2)**2
    b = s1**4 / (n1**2 * (n1-1)) + s2**2 / (n2**2 * (n2-1))
    return a/b






def local_median(img, size):

    def function(x):

        valids = np.isfinite(x)

        if np.any(valids):
            return np.nanmedian(x)
        else:
            return np.nan

    return ndi.generic_filter(img, function, size)



def local_mean(img, size, kernel_type='neighborhood', var_rand=1.0):
    """
    Compute local mean, variance reduction and number of valid pixels using
    different kernel_types:
    
    - neighborhood (size: 1,5,9,13,25)
    - gaussian (size is standard deviation)
    - uniform
    
    """
    if kernel_type == 'gaussian':
        footprint = gaussian_kernel(sigma=size)
    
    elif kernel_type == 'uniform':
        footprint = np.ones((size, size))
        
    elif kernel_type == 'neighborhood':
        if size == 1:
            return img.copy(), np.ones_like(img)

        elif size == 5:
            footprint = np.array([
                [0, 1, 0],
                [1, 1, 1],
                [0, 1, 0]
            ])
        elif size == 13:
            footprint = np.array([
                [0, 0, 1, 0, 0],
                [0, 1, 1, 1, 0],
                [1, 1, 1, 1, 1],
                [0, 1, 1, 1, 0],
                [0, 0, 1, 0, 0],
            ])
        elif size == 37:
            footprint = np.array([
                [0,0,1,1,1,0,0],
                [0,1,1,1,1,1,0],
                [1,1,1,1,1,1,1],
                [1,1,1,1,1,1,1],
                [1,1,1,1,1,1,1],
                [0,1,1,1,1,1,0],
                [0,0,1,1,1,0,0],
            ])
        elif size in [9,25,49,81,121]:
            size = int(np.sqrt(size))
            footprint = np.ones((size,size))
        else:
            raise ValueError('size of neighborhood %d not supported' % size)
            
    else:
        raise ValueError('"kernel_type" needs to be "gaussian", "uniform" or "neighborhood"')

    # only compute if any values in across-track direction
    domain = np.any(img, axis=1)

    # normalize footprint
    footprint = footprint / footprint.sum()
    
    # keep dtype of image
    footprint = footprint.astype(img.dtype)

    # compute mean, variance and number of valid pixels
    mean = np.full(img.shape, np.nan, dtype=img.dtype)
    var = np.full(img.shape, np.nan, dtype=img.dtype)
    n = np.full(img.shape, np.nan, dtype=int)
   
    mean[domain,:] = ndi.generic_filter(img[domain,:], function=weighted_mean,
                                        size=footprint.shape, mode='constant', cval=np.nan,
                                        extra_arguments=(footprint.flatten(),))

    var[domain, :] = ndi.generic_filter(img[domain,:], function=weighted_mean_var,
                                        size=footprint.shape, mode='constant', cval=np.nan,
                                        extra_arguments=(footprint.flatten(), var_rand))
    n[domain,:] = ndi.generic_filter(
            np.isfinite(img).astype(int)[domain,:], np.count_nonzero,
            footprint=(footprint != 0), mode='constant', cval=0
    )

    return mean, var, n






